/* Copyright (c) 1997-2003 Miller Puckette.
* For information on usage and redistribution, and for a DISCLAIMER OF ALL
* WARRANTIES, see the file, "LICENSE.txt," in this distribution.

iemlib1 written by Thomas Musil (c) IEM KUG Graz Austria 2000 - 2003 */

#ifdef NT
#pragma warning( disable : 4244 )
#pragma warning( disable : 4305 )
#endif


#include "m_pd.h"
#include "iemlib.h"
#include <math.h>
#include <stdio.h>
#include <string.h>

/* ------------------------ iem_cot~ ----------------------------- */

float *iem_iem_cot_table_cos=(float *)0L;
float *iem_iem_cot_table_sin=(float *)0L;

static t_class *sigiem_cot_class;

typedef struct _sigiem_cot
{
    t_object x_obj;
    float x_msi;
} t_sigiem_cot;

static void *sigiem_cot_new(void)
{
    t_sigiem_cot *x = (t_sigiem_cot *)pd_new(sigiem_cot_class);

    outlet_new(&x->x_obj, gensym("signal"));
    x->x_msi = 0;
    return (x);
}

static t_int *sigiem_cot_perform(t_int *w)
{
    t_float *in = (t_float *)(w[1]);
    t_float *out = (t_float *)(w[2]);
    t_float hout;
    int n = (int)(w[3]);
    float *ctab = iem_iem_cot_table_cos, *stab = iem_iem_cot_table_sin;
    float *caddr, *saddr, cf1, cf2, sf1, sf2, frac;
    double dphase;
    int normhipart;
    int32 mytfi;
    union tabfudge tf;
    
    tf.tf_d = UNITBIT32;
    normhipart = tf.tf_i[HIOFFSET];

#if 0	    	/* this is the readable version of the code. */
    while (n--)
    {
	if(*in < 0.0000001)
	    *in = 0.0000001;
	else if(*in > 0.9999999)
            *in = 0.9999999;
    	dphase = (double)(*in++ * (float)(COSTABSIZE)) + UNITBIT32;
	tf.tf_d = dphase;
	mytfi = tf.tf_i[HIOFFSET] & (COSTABSIZE-1);
        saddr = stab + (mytfi);
	caddr = ctab + (mytfi);
	tf.tf_i[HIOFFSET] = normhipart;
	frac = tf.tf_d - UNITBIT32;
	sf1 = saddr[0];
	sf2 = saddr[1];
        cf1 = caddr[0];
	cf2 = caddr[1];
	*out++ = (cf1 + frac * (cf2 - cf1))/(sf1 + frac * (sf2 - sf1));
    }
#endif
#if 1	    	/* this is the same, unwrapped by hand. prolog beg*/
        n /= 4;
        if(*in < 0.0000001)
	    *in = 0.0000001;
	else if(*in > 0.9999999)
            *in = 0.9999999;
    	dphase = (double)(*in * (float)(COSTABSIZE)) + UNITBIT32;
        tf.tf_d = dphase;
        mytfi = tf.tf_i[HIOFFSET] & (COSTABSIZE-1);
        saddr = stab + (mytfi);
	caddr = ctab + (mytfi);
	tf.tf_i[HIOFFSET] = normhipart;        
        in += 4;                               /*prolog end*/
    while (--n)
    {
        if(*in < 0.0000001)
	    *in = 0.0000001;
	else if(*in > 0.9999999)
            *in = 0.9999999;
    	dphase = (double)(*in * (float)(COSTABSIZE)) + UNITBIT32;
	    frac = tf.tf_d - UNITBIT32;
	tf.tf_d = dphase;
	    sf1 = saddr[0];
	    sf2 = saddr[1];
            cf1 = caddr[0];
	    cf2 = caddr[1];
        mytfi = tf.tf_i[HIOFFSET] & (COSTABSIZE-1);
        saddr = stab + (mytfi);
	caddr = ctab + (mytfi);
	    hout = (cf1 + frac * (cf2 - cf1))/(sf1 + frac * (sf2 - sf1));
	    *out++ = hout;
	    *out++ = hout;
	    *out++ = hout;
	    *out++ = hout;
            in += 4;
	tf.tf_i[HIOFFSET] = normhipart;

    }/*epilog beg*/
	    frac = tf.tf_d - UNITBIT32;
            sf1 = saddr[0];
	    sf2 = saddr[1];
            cf1 = caddr[0];
	    cf2 = caddr[1];
	    hout = (cf1 + frac * (cf2 - cf1))/(sf1 + frac * (sf2 - sf1));
            *out++ = hout;
	    *out++ = hout;
	    *out++ = hout;
	    *out++ = hout;
            /*epilog end*/
#endif
    return (w+4);
}

static void sigiem_cot_dsp(t_sigiem_cot *x, t_signal **sp)
{
    dsp_add(sigiem_cot_perform, 3, sp[0]->s_vec, sp[1]->s_vec, sp[0]->s_n);
}

static void sigiem_cot_maketable(void)
{
    int i;
    float *fp, phase, fff, phsinc = 0.5*3.141592653 / ((float)COSTABSIZE);
    union tabfudge tf;

    if(!iem_iem_cot_table_sin)
    {
	iem_iem_cot_table_sin = (float *)getbytes(sizeof(float) * (COSTABSIZE+1));
	for(i=COSTABSIZE+1, fp=iem_iem_cot_table_sin, phase=0; i--; fp++, phase+=phsinc)
	    *fp = sin(phase);
    }
    if(!iem_iem_cot_table_cos)
    {
	iem_iem_cot_table_cos = (float *)getbytes(sizeof(float) * (COSTABSIZE+1));
	for(i=COSTABSIZE+1, fp=iem_iem_cot_table_cos, phase=0; i--; fp++, phase+=phsinc)
            *fp = cos(phase);
    }
    tf.tf_d = UNITBIT32 + 0.5;
    if((unsigned)tf.tf_i[LOWOFFSET] != 0x80000000)
	bug("iem_cot~: unexpected machine alignment");
}

void sigiem_cot_setup(void)
{
    sigiem_cot_class = class_new(gensym("iem_cot~"), (t_newmethod)sigiem_cot_new, 0,
				 sizeof(t_sigiem_cot), 0, 0);
    class_addcreator((t_newmethod)sigiem_cot_new, gensym("icot~"), 0);
    CLASS_MAINSIGNALIN(sigiem_cot_class, t_sigiem_cot, x_msi);
    class_addmethod(sigiem_cot_class, (t_method)sigiem_cot_dsp, gensym("dsp"), 0);
    sigiem_cot_maketable();
    class_sethelpsymbol(sigiem_cot_class, gensym("iemhelp/help-iem_cot~"));
}

